# -*- coding: utf-8 -*-
"""
Input/output functions for interaction with user's script
=========================================================

Parsers and formatters to be used with user's input files or cli arguments and expected outputs.

"""
