"""
Benchmark Assessments definition
=================================
"""

from .averagerank import AverageRank
from .averageresult import AverageResult
from .base import BaseAssess

__all__ = ["BaseAssess", "AverageRank", "AverageResult"]
