# -*- coding: utf-8 -*-
"""
Experiment Version Control System
=================================

Organize related experiments inside a tree-structure using adapters to maintain
compatibility between them.
"""
